from dataclasses import dataclass, field
from j2pipeline.template import assemble
from j2pipeline.tcp import client
from typing import Callable, NoReturn

@dataclass(slots=True)
class Prompt[T]:
    path: str
    process: Callable[[str], T] = field(default=lambda response: response)
    auto_upper: bool = field(default=True)
    __callback: Callable[[str], NoReturn] = field(init=False)
    __response: str | None = field(default=None, init=False)

    def __post_init__(self) -> None:
        def callback(response: str) -> NoReturn:
            self.__response = response
            exit(code=0)
        self.__callback = callback

    def __call__(self, **subs: str) -> T:
        self.__response = None
        if self.auto_upper:
            subs = { key.upper(): value for key, value in subs.items() }
        prompt: str = assemble(path=self.path, subs=subs)
        with client(on_receive=self.__callback) as clt:
            clt.send(prompt)
        while True:
            if self.__response is not None:
                return self.process(self.__response)


from dataclasses import dataclass, field
from os import listdir
from j2pipeline.types.prompt import Prompt
from typing import Callable

@dataclass(slots=True)
class Pipeline[T]:
    base_path: str
    reorder: Callable[[list[str]], list[str]] = field(default=lambda prompts: prompts)
    process: dict[str, Callable[[str], str]] = field(default_factory=dict)
    final_process: Callable[[str], T] = field(default=lambda result: result)
    extension: str = field(default='j2')
    __names: list[str] = field(init=False)

    def __post_init__(self) -> None:
        self.__names = self.reorder([name[:-len(f'.{self.extension}')]
                                     for name
                                     in listdir(path=self.base_path)])
        if '*' not in self.process.keys():
            self.process.update({ '*': lambda response: response })
    
    def __call__(self, prompt: str) -> T:
        result: str = prompt
        for name in self.__names:
            prompt: Prompt = Prompt(path=f'{self.base_path}\\{name}.{self.extension}',
                                    process=self.process.get(name, self.process['*']))
            result: str = prompt(PROMPT=result)
        return self.final_process(result)


from j2pipeline.template.load import load_template
from j2pipeline.template.vars import get_vars, substitute

def assemble(path: str, subs: dict[str, str]) -> str:
    template: str = load_template(path=path)
    vars: dict[str, str] = get_vars(j2=template)
    return substitute(j2=template, subs=subs, vars=vars)


from re import findall

def get_vars(j2: str) -> dict[str, str]:
    pattern: str = r'(?<!\\){%[ \n]*[_a-zA-Z][_a-zA-Z0-9]*[ \n]*[^\\]%}'
    matches: list[str] = findall(pattern=pattern, string=j2)
    return { var[2:-2].strip(): var for var in matches }

def substitute(j2: str,
                subs: dict[str, str],
                vars: dict[str, str]) -> str:
    result: str = j2
    for name, sub in subs.items():
        result = result.replace(vars[name], sub)
    return result


from socket import socket as Socket
from threading import Thread
from dataclasses import dataclass, field
from typing import Callable, NoReturn

@dataclass(slots=True)
class Client:
    socket: Socket
    on_receive: Callable[[str], NoReturn]
    receive_thread: Thread | None = field(default=None, init=False)

    def __post_init__(self) -> None:
        def receive() -> None:
            while True:
                try:
                    size: int = int(self.socket.recv(16).decode(encoding='UTF-8').strip())
                    message: str = self.socket.recv(size).decode(encoding='UTF-8')
                    self.on_receive(message)
                except Exception as e:
                    print(e)
        self.receive_thread = Thread(target=receive, daemon=True)
        self.receive_thread.start()

    def send(self, message: str) -> None:
        data: bytes = message.encode(encoding='UTF-8')
        size: bytes = f'{len(data):16}'.encode(encoding='UTF-8')
        self.socket.send(size)
        self.socket.send(data)

    def join(self, timeout: float | None = None) -> None:
        self.receive_thread.join(timeout=timeout)

    def close(self) -> None:
        self.socket.close()


from socket import socket as Socket, AddressFamily, SocketKind
from contextlib import contextmanager
from typing import Generator, Callable
from j2pipeline.tcp.client import Client

@contextmanager
def client(on_receive: Callable[[str], None],
           timeout: float | None = None,
           host: str = 'localhost',
           port: int = 50027) -> Generator[Client, None, None]:
    socket: Socket = Socket(family=AddressFamily.AF_INET,
                            type=SocketKind.SOCK_STREAM)
    socket.connect((host, port))
    client: Client = Client(socket=socket, on_receive=on_receive)
    try:
        yield client
    finally:
        client.join(timeout=timeout)
        client.close()


import setuptools

with open(file='README.md', mode='r', encoding='UTF-8') as readme_file:
    readme: str = readme_file.read()

setuptools.setup(
    name='j2pipeline',
    version='0.3.0',
    description='Create LLM pipelines with ease!',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/yaz008/j2pipeline.git',
    author='Emelianov Artem',
    author_email='yaz008.yaz008@yandex.ru',
    license='MIT',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.12",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    include_package_data=True
)